/** \file   dd_io_kf.h
 *  \brief  IO functions for data files in Knuth format (for ZDDs)
 *  \author Takahisa Toda
 *  \see pp.206-207 in "The Art of Computer Programming, Volume 4a (2011)" by Knuth
 *  \par Knuth Format DEFINITION
 *  - A sequential list of branch instructions I_{m}, I_{m-1},...,I_{1}
 *  - I_{k} has the form n_{k}: (~v_{k}?l_{k}:h_{k}), which represents an internal ZDD node.
 *  - n_{k}: the unique index (hexadecimal number) assigned to a ZDD node,
 *  - v_{k}: the label (decimal number) of a ZDD node,
 *  - l_{k}: the unique index (hexadecimal number) assigned to the LO-child,
 *  - h_{k}: the unique index (hexadecimal number) assigned to the HI-child.
 *  - The node indices of the top and the bottom nodes must be 1 and 0, respectively.
 *  \par An example of data file in a Knuth format is shown below.
 * \code
 * ------EXAMPLE------
 * 2: (~10?1:1)
 * 3: (~9?0:2)
 * 4: (~8?0:1)
 * 5: (~7?3:4)
 * 6: (~6?0:4)
 * 7: (~5?5:6)
 * 8: (~9?0:1)
 * 9: (~8?0:8)
 * a: (~4?7:9)
 * b: (~7?0:1)
 * c: (~4?0:b)
 * d: (~2?a:c)
 * e: (~10?0:1)
 * f: (~3?0:e)
 * 10: (~2?0:f)
 * 11: (~1?d:10)
 * -------------------
 * \endcode
 * This data corresponds to the following data file.
 * \code
 * ------EXAMPLE------
 * 2 4 7 
 * 7 8 
 * 9 
 * 9 10 
 * 4 8 9 
 * 1 2 3 10 
 * 5 6 8 
 * -------------------
 * \endcode
 */
#ifndef DD_IO_KF_H
#define DD_IO_KF_H

#include <stdint.h>

#include "dd_wrapper.h"

/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/

extern int getfileinfo_kf(size_t *pnrow, size_t *pmaxval, FILE *in);
extern my_zddp zimport_kf(size_t nrow, FILE *in);
extern int zexport_kf(size_t maxval, my_zddp f, FILE *out);

#endif /*DD_IO_KF_H*/
