/** \file   dd_io_setfam.c
 *  \brief  Implementation of IO functions for data files that represent set families.
 *  \author Takahisa Toda
 *  \note See: T.Toda, Fast Compression of Large-scale Hypergraphs for Solving Combinatorial Problems, in Proc. of Sixteenth International Conference on Discovery Science (DS2013), LNAI 8140, LNAI 8140, pp. 281--293, Singapore (2013).
 */
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "my_def.h"
#include "ssort1.h"
#include "dd_wrapper.h"
#include "my_io_setfam.h"
#include "dd_io_setfam.h"

static my_zddp bpconstruct_rec(int **s, size_t n, int depth);
static inline size_t my_bsearch(int **s, size_t left, size_t right, int v, int depth);

static size_t recdepth = 0; //!< recursion depth

/**  \brief Build a ZDD from input data (in a binary partitionining method after multikey Quicksort).
 *  \param  nrow        the number of rows in a data file
 *  \param  total_nent  the total number of entries in a data file
 *  \param  in          pointer to a dafa file
 *  \return the built ZDD
 *  \see zexport_setfam
 */
my_zddp zimport_setfam(size_t nrow, size_t total_nent, FILE *in)
{
  struct setfam *sf = in2setfam(nrow, total_nent, in);
  ENSURE_TRUE(sf != NULL);
  ssort1main(sf->elem, nrow);
  assert(recdepth == 0);
  my_zddp f = bpconstruct_rec(sf->elem, nrow, 0);
  ENSURE_TRUE(f != ZDD_NULL);
  assert(recdepth == 0);
  destroy_setfam(sf);
  return f;
}

/** \brief Build a ZDD from a set family in a binary partitionining method.
 * \param s      s[i] corresponds to the i-th set
 * \param n      the number of sets
 * \param depth  the position of a set to be examined.
 * \return a root node
 */
static my_zddp bpconstruct_rec(int **s, size_t n, int depth)
{ 
  size_t pos; int v; my_zddp lo, hi;
  if (n == 0)  return zdd_bot();
  v = s[0][depth];
  if (v == INFTY) return zdd_top();
  INC_RECDEPTH(recdepth);  
  if(s[n-1][depth] == v) {pos = n;}
  else {
    for(pos = 1; pos < n && s[pos][depth] == v; pos *= 2) ;
    if(pos < n) pos = my_bsearch(s, pos/2, pos, v, depth);
    else  pos = my_bsearch(s, pos/2, n-1, v, depth);
  }
  hi = bpconstruct_rec(s, pos, depth+1);
  lo = bpconstruct_rec(s+pos, n-pos, depth);
  my_zddp f = zdd_unique(v, lo, hi);
  ENSURE_TRUE_MSG(f != ZDD_NULL, "ZDD operation failed");
  DEC_RECDEPTH(recdepth);
  return f;
}


/** \brief Find the first position with depth-th value not equal to v (by a binary search).
 * \param s      s[i] corresponds to the i-th row in a data file
 * \param left   the leftend position
 * \param right  the rightend position
 * \param v      value
 * \param depth  the position of rows to be examined.
 * \return the found position.
 * \note assume that (v ==) s[left][depth] <= s[left+1][depth] <= ... <= s[right][depth] (!= v)
 */
static inline size_t my_bsearch(int **s, size_t left, size_t right, int v, int depth)
{
  while(1) {
    assert(s[left][depth] == v); assert(s[right][depth] != v); assert(left < right);
    if(right - left <= 1) return right;
    size_t mid = (left + right) / 2;
    if(s[mid][depth] == v) left = mid;
    else right = mid;
  }
}


/**  \brief  Build a ZDD from input data (in a naive method)
 *  \param  nrow        the number of rows
 *  \param  total_nent  the total number of entries in a data file
 *  \param  in          pointer to a dafa file
 *  \return a root node
 */
my_zddp naive_zimport_setfam(size_t nrow, size_t total_nent, FILE *in)
{
  struct setfam *sf = in2setfam(nrow, total_nent, in);
  ENSURE_TRUE(sf != NULL);

  my_zddp f = zdd_bot();
  for(size_t i = 0; i < nrow; i++) {
    my_zddp g = zdd_top();
    int j = 0;
    int *x = sf->elem[i];
    while(x[j] != INFTY) j++;
    while(--j >= 0) {
      g = zdd_unique(x[j], zdd_bot(), g);
      ENSURE_TRUE_MSG(g != ZDD_NULL, "ZDD operation failed");
    }
    f = zdd_union(f, g);
    ENSURE_TRUE_MSG(f != ZDD_NULL, "ZDD operation failed");
  }

  destroy_setfam(sf);  
  return f;
}


/**  \brief  Build a ZDD from input data (in a naive method after sorting)
 *  \param  nrow        the number of rows
 *  \param  total_nent  the total number of entries in a data file
 *  \param  in          pointer to a dafa file
 *  \return a root node
 */
my_zddp sort_naive_zimport_setfam(size_t nrow, size_t total_nent, FILE *in)
{
  struct setfam *sf = in2setfam(nrow, total_nent, in);
  ENSURE_TRUE(sf != NULL);

  ssort1main(sf->elem, nrow); // sorting

  my_zddp f = zdd_bot();
  for(size_t i = 0; i < nrow; i++) {
    my_zddp g = zdd_top();
    int j = 0;
    int *x = sf->elem[i];
    while(x[j] != INFTY) j++;
    while(--j >= 0) {
      g = zdd_unique(x[j], zdd_bot(), g);
      ENSURE_TRUE_MSG(g != ZDD_NULL, "ZDD operation failed");
    }
    f = zdd_union(f, g);
    ENSURE_TRUE_MSG(f != ZDD_NULL, "ZDD operation failed");
  }

  destroy_setfam(sf);  
  return f;
}


/**  \brief Export ZDD into a data file.
 *  \param maxval   the maximum value of an element
 *  \param  f     ZDD, which are not constant
 *  \param  out   pointer to a dafa file
 *  \return MY_SUCCESS if successful; otherwise, MY_FAILURE.
 *  \see zimport_setfam
 */
int zexport_setfam(size_t maxval, my_zddp f, FILE *out)
{

  if(zdd_isconst(f)) {
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_WARN
    fprintf(stderr, "warning: nothing is exported because ZDD is constant.\n");
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_WARN*/
    return MY_FAILURE;
  }
  
  int res = fseek(out, 0L, SEEK_SET);
  ENSURE_SUCCESS(res);

  my_zddp *br = (my_zddp*)malloc(sizeof(my_zddp)*(maxval+1));
  ENSURE_TRUE_MSG(br != NULL, "memory allocation failed");
  my_zddp *bp = br;
  size_t *hi = (size_t*)malloc(sizeof(size_t)*(maxval+1));
  ENSURE_TRUE_MSG(hi != NULL, "memory allocation failed");
  size_t *hp = hi;

  while(1) {
    while(f != zdd_top()) {
      my_zddp fl = zdd_LO(f);
      my_zddp fh = zdd_HI(f);
      fprintf(out, "%zu ", zdd_label(f));
      assert(hp < hi+maxval+1);
      *hp = zdd_label(f); hp++;
      if(fl != zdd_bot()) {assert(bp < br+maxval+1); *bp = f; bp++;}
      f = fh;
    }
    fprintf(out, "\n");
    if((bp--) <= br) break;
    size_t max = zdd_label(*bp);
    for(hp = hi; *hp < max; hp++) {
      assert(hp < hi+maxval+1);
      fprintf(out, "%zu ", *hp);
    }
    f = zdd_LO(*bp);
    assert(f != zdd_bot());
  }
  
  free(br); free(hi);
  br = NULL; hi = NULL;
  return MY_SUCCESS;
}

