/** \file   my_cache.c
 *  \brief  Implementation of a simple cache for binary operations
 *  \author Takahisa Toda
 */
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "my_cache.h"
#include "my_def.h"

enum CACHE_POSKEY {
  CACHE_POS_OPID = 0,
  CACHE_POS_OPERAND1,
  CACHE_POS_OPERAND2,
  CACHE_POS_RESULT,
  CACHE_ENTRY_SIZE, /* if you want to add a new field in a cache entry, insert it before CACHE_ENTRY_SIZE.*/
};


/** \brief  Create a cache.
 *  \param  num the number of cache entries
 *  \return pointer to a cache if successful; otherwise, NULL.
 *  \see    cache_destroy, cache_enlarge
 */
struct my_cache *cache_create(size_t num)
{
  ENSURE_TRUE(num != 0);
  struct my_cache *ca = (struct my_cache*) malloc(sizeof(struct my_cache));
  ENSURE_TRUE_MSG(ca != NULL,"memory allocation failed");

  ca->nent = num;
  const size_t len = num*CACHE_ENTRY_SIZE;
  ca->entry = (size_t*) malloc(sizeof(size_t) * len);
  ENSURE_TRUE_MSG(ca->entry != NULL, "memory allocation failed");
  for(size_t i = 0; i < len; i++) ca->entry[i] = 0;

  return ca;
}


/** \brief Free memory area used by a cache.
 *  \param ca pointer to a cache
 *  \see   cache_create, cache_enlarge
 */
void cache_destroy(struct my_cache *ca)
{
  ENSURE_TRUE(ca != NULL);
  free(ca->entry); ca->entry = NULL;
  free(ca); ca = NULL;
}

/** \brief  Calculate a hash value.
 *  \param  id   operation id (>0)
 *  \param  k1   operand
 *  \param  k2   operand
 *  \param  num  the total number of cache entries
 *  \return the calculated hash value, which must be less than the number of cache entries.
 *  \note   The position of a cache entry can be computed by multipling a hash value by CACHE_ENTRY_SIZE.
 */
static inline size_t my_hash(size_t id, size_t k1, size_t k2, size_t num)
{
  assert(num != 0);
  return (id*MY_PRIME(0) + k1*MY_PRIME(1) + k2*MY_PRIME(2))%num;
}


/** \brief  Register a new entry.
 *  \param  id   operation id (> 0)
 *  \param  k1   operand
 *  \param  k2   operand
 *  \param  r    result
 *  \see    cache_search
 *  \note   Overwrite if the position is already filled.
 */
void cache_add(size_t id, size_t k1, size_t k2, size_t r, struct my_cache *ca)
{
  assert(id != 0);
  const size_t basepos = CACHE_ENTRY_SIZE * my_hash(id, k1, k2, ca->nent);
  ca->entry[basepos + CACHE_POS_OPID]     =  id;
  ca->entry[basepos + CACHE_POS_OPERAND1] =  k1;
  ca->entry[basepos + CACHE_POS_OPERAND2] =  k2;
  ca->entry[basepos + CACHE_POS_RESULT]   =  r;
}


/** \brief  Find a cache entry. If found, copy the operation result to pr.
 *  \param  id   operation id (> 0)
 *  \param  k1   operand
 *  \param  k2   operand
 *  \param  pr   pointer to the operation result
 *  \param  ca   pointer to a cache
 *  \return MY_SUCCESS if found; otherwise, MY_FAILURE.
 *  \see    cache_add
 *  \note
 *  - result is not copied if pr == NULL.
 */
int cache_search(size_t id, size_t k1, size_t k2, size_t *pr, const struct my_cache *ca)
{
  assert(id != 0);
  const size_t basepos = CACHE_ENTRY_SIZE * my_hash(id, k1, k2, ca->nent);
  if(ca->entry[basepos+CACHE_POS_OPID] == id 
     && ca->entry[basepos+CACHE_POS_OPERAND1] == k1 
     && ca->entry[basepos+CACHE_POS_OPERAND2] == k2) {
    if(pr != NULL) *pr = ca->entry[basepos+CACHE_POS_RESULT];
    return MY_SUCCESS;
  } else {return MY_FAILURE;}
}


/* \brief Enlarge cache size.
 * \param ca pointer to a cache
 * \see cache_create, cache_destroy
 */
void cache_enlarge(struct my_cache *ca)
{
  const size_t oldsiz = ca->nent; 
  const size_t newsiz = oldsiz << 2; const size_t newlen = CACHE_ENTRY_SIZE*newsiz;
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO
  printf("cache resizing:%zu -> %zu\n", oldsiz, newsiz);
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO*/
  size_t *oldmem = ca->entry;
  size_t *newmem = (size_t*) malloc(sizeof(size_t) * newlen);
  ENSURE_TRUE_MSG(newmem != NULL, "memory allocation failed");
  for(size_t i = 0; i < newlen; i++) newmem[i] = 0;
  for(size_t i = 0; i < oldsiz; i++) {
    const size_t oldbasepos = CACHE_ENTRY_SIZE * i;
    const size_t newbasepos = CACHE_ENTRY_SIZE * my_hash(oldmem[oldbasepos+CACHE_POS_OPID], 
							 oldmem[oldbasepos+CACHE_POS_OPERAND1], 
							 oldmem[oldbasepos+CACHE_POS_OPERAND2], 
							 newsiz);
    newmem[newbasepos+CACHE_POS_OPID]     = oldmem[oldbasepos+CACHE_POS_OPID];
    newmem[newbasepos+CACHE_POS_OPERAND1] = oldmem[oldbasepos+CACHE_POS_OPERAND1];
    newmem[newbasepos+CACHE_POS_OPERAND2] = oldmem[oldbasepos+CACHE_POS_OPERAND2];
    newmem[newbasepos+CACHE_POS_RESULT]   = oldmem[oldbasepos+CACHE_POS_RESULT];
  }
  free(oldmem); oldmem = NULL;
  ca->nent  = newsiz;
  ca->entry = newmem;
}


