/** \file   my_io.h
 *  \brief  IO functions for data files that represet set families.
 *  \author Takahisa Toda
 *  \par Data File Format
 * - Entries must be sorted in increasing order and separated by the character indicated by the macro SF_SEPARATOR.
 * - No duplicated entries appear in the same row.
 * - Each entry must be a nonzero positive number.
 * - (empty line is allowed).
 * - The bytesize of a line in a datafile must be less than the value indicated by the macro MAX_LINESIZ.
 *
 * \code
 * ------EXAMPLE------
 * 2 4 7 
 * 7 8 
 * 9 
 * 9 10 
 * 4 8 9 
 * 1 2 3 10 
 * 5 6 8 
 * -------------------
 * \endcode
 *
 */
#ifndef MY_IO_SETFAM_H
#define MY_IO_SETFAM_H

#define SF_SEPARATOR ' '
#define MAX_LINESIZ  (1UL << 17) //!< the maximum bytesize of a line in a data file

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

/** \brief an explicite data represention of a set family
 *  \note
 *  - Let sf point to this data structure.
 *  - sf->elem[i-1][j-1] is the j-th element in the i-th set. 
 *  - sf->elem[i-1][k] == INFTY if the i-th set contains exactly k elements.
 */
struct setfam {
  size_t nset;    //!< the number of sets with repetition (i.e., the number of rows in a data file)
  size_t memsize; //!< the length of the array mem
  int **elem;
  int *mem;
};

/*---------------------------------------------------------------------------*/
/* Function prototypes                                                       */
/*---------------------------------------------------------------------------*/
extern int getfileinfo(size_t *pnrow, size_t *pmaxval, size_t *ptotal_nent, FILE *in);
extern struct setfam *in2setfam(size_t nrow, size_t total_nent, FILE *in);
extern void destroy_setfam(struct setfam *sf);

#endif /*MY_IO_SETFAM_H*/
