/** \file   my_random.h
 *  \brief  functions to manipulate a random number generator.
 *  \author Takahisa Toda
 *  \note If the macro NGSL is defined, then use a default random number generator. Otherwise, use a random number generator provided by GNU GSL.
 */
#ifndef MY_RANDOM_H
#define MY_RANDOM_H

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>

#include "my_def.h"

#ifndef NGSL
/*---------------------------------------------------------------------------*/
/* GNU GSL                                                                   */
/*---------------------------------------------------------------------------*/
#include <gsl/gsl_math.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

typedef gsl_rng my_rng;

extern my_rng *g_rng;

/** \brief Calculate a random number.
 *  \return a random number r with 0 <= r < n.
 *  \see my_initrand
 *  \note
 *  - Note: Before using this function, execute my_initrand() once.
 */
static inline size_t my_rand(size_t n)
{
  assert(g_rng != NULL);
  assert(n <= gsl_rng_max(g_rng));
  return (size_t)gsl_rng_get(g_rng)%n;
}


/** \brief Create and initialize a random number generator.
 *  \see my_rand 
 */
static inline void my_initrand(void)
{
  if (g_rng == NULL) {
    g_rng = gsl_rng_alloc(gsl_rng_default);
    ENSURE_TRUE_MSG(g_rng != NULL, "a random number generator creation failed");
  } else {
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_WARN
    fprintf(stderr, "warning: random number generator (%s) already initialized.\n", gsl_rng_name(g_rng));
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_WARN*/
  }

  const time_t now = time(NULL);
  assert(now != (time_t) -1);
  gsl_rng_set(g_rng, (unsigned long int) now);

#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO
  printf("\ngenerator type:\t%s\n", gsl_rng_name(g_rng));
  printf("first value:\t%lu\n", gsl_rng_get(g_rng)); 
  printf("max value:\t%lu\n",  gsl_rng_max(g_rng));
  printf("seed:\t\t%lu\n", (unsigned long int)now);
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO*/
}

/** \brief Free memory area
 *  \see my_initrand 
 */
static inline void my_freerand(void)
{
  if (g_rng != NULL) {
    gsl_rng_free(g_rng);
    g_rng = NULL;
  } else {
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_WARN
    fprintf(stderr, "warning: random number generator already freed.\n");
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_WARN*/
  }
}

#else /*NGSL*/
/*---------------------------------------------------------------------------*/
/* DEFAULT RANDOM NUMBER GENERATOR                                           */
/*---------------------------------------------------------------------------*/

/** \brief Calculate a random number.
 *  \return a random number r with 0 <= r < n.
 *  \see my_initrand
 *  \note
 *  - Note: Before using this function, execute my_initrand() once.
 */
static inline size_t my_rand(size_t n)
{
  return (size_t)random()%n;
}


/** \brief Create and initialize a random number generator.
 *  \see my_rand 
 */
static inline void my_initrand(void)
{
  const time_t now = time(NULL);
  assert(now != (time_t) -1);
  srandom((unsigned int) now);
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO
  printf("\ngenerator type:\tdefault\n");
  printf("first value:\t%d\n", rand()); 
  printf("max value:\t%d\n",  RAND_MAX);
  printf("seed:\t\t%u\n", (unsigned int) now);
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO*/
}


/** \brief Free memory area
 *  \see my_initrand 
 */
static inline void my_freerand(void)
{
}
#endif  /*NGSL*/

#endif /*MY_RANDOM_H*/
