/** \file   unzcomp.c
 *  \brief  Decompression of ZDDs into data files in Knuth form
 *  \author Takahisa Toda
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "my_def.h"
#include "my_random.h"
#include "my_cache.h"
#include "dd_io_setfam.h"
#include "dd_io_kf.h"

#define PROGNAME "UNZCOMP"//!< the name of this program
#define PROGVERS "1.1.0"  //!< the current version

#ifdef SELECT_CUDD
DdManager *dd_mgr = NULL; //!< ZDD manager for CUDD
#endif /*SELECT_CUDD*/

#ifndef NGSL
my_rng *g_rng = NULL; //!< random number generator for GSL
#endif /*NGSL*/

#define PRINT_USAGE(p) do{fprintf(stderr, "Usage: %s input-file [output-file]\n", (p));} while(0)


int main(int argc, char *argv[])
{   
  int res = MY_SUCCESS;
  char *infile = NULL;
  char *outfile = NULL;

  /*** INPUT ARGUMENT ***/  
  for(int i = 1; i < argc; i++) {
    if(argv[i][0] == '-') {
      switch (argv[i][1]){
      case 'v':
	fprintf(stderr, "%s\tversion %s\n\n", PROGNAME, PROGVERS);
	return MY_SUCCESS;
      case '?': case 'h': default:
	PRINT_USAGE(argv[0]); return  MY_SUCCESS;
      }
    } else {
      if(infile == NULL) {infile = argv[i];}
      else if(outfile == NULL) {outfile = argv[i];}
      else {PRINT_USAGE(argv[0]); return  MY_SUCCESS;}
    }
  }
  if(infile == NULL) {PRINT_USAGE(argv[0]); return  MY_SUCCESS;}
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO
  printf("%s\t\tversion %s\n", PROGNAME, PROGVERS);
  printf("package\t\t%s\n", ZDD_PACKAGE);
  printf("input\t\t%s\n", infile);
  if(outfile != NULL) printf("output\t\t%s\n", outfile);
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO*/

  my_initrand(); // initialize random number generator
  FILE *in = fopen(infile, "r");
  ENSURE_TRUE_MSG(in != NULL, "file open failed\n");

  /*** INITIALIZE ZDD MANAGER ***/
  size_t nrow, maxval;
  res = getfileinfo_kf(&nrow, &maxval, in);
  ENSURE_SUCCESS(res);
  // adjust CACHE_DEFAULT_SIZE if necessary.
  zdd_init(maxval, CACHE_DEFAULT_SIZE);

  /*** IMPORT ZDD ***/  
  my_zddp f;
  clock_t startingtime = my_clock();
  f = zimport_kf(nrow, in);
  print_elapsed_time(startingtime, "IMPORT");    
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO
    printf("|zdd|\t\t%zu\n", zdd_size(f));
    printf("#sets\t\t%zu\n\n", zdd_card(f));
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO*/


  /*** EXPORT ZDD ***/  
  if(outfile != NULL) {
    FILE *out = fopen(outfile, "w");
    startingtime = my_clock();
    zexport_setfam(maxval, f, out);
    print_elapsed_time(startingtime, "EXPORT");
    fclose(out);
  }

  zdd_quit(); my_freerand();
  fclose(in);

  return MY_SUCCESS;
}


