/** \file   zcomp.c
 *  \brief  Compression of Hypergraphs into ZDDs
 *  \author Takahisa Toda
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "my_def.h"
#include "my_random.h"
#include "my_cache.h"
#include "dd_io_setfam.h"
#include "dd_io_kf.h"

#define PROGNAME "ZCOMP"//!< the name of this program
#define PROGVERS "1.1.0" //!< the current version

#ifdef SELECT_CUDD
DdManager *dd_mgr = NULL; //!< ZDD manager for CUDD
#endif /*SELECT_CUDD*/

#ifndef NGSL
my_rng *g_rng = NULL; //!< random number generator for GSL
#endif /*NGSL*/

#define PRINT_USAGE(p) do{fprintf(stderr, "Usage: %s [option] input-file [output-file]\n", (p)); \
    fprintf(stderr, "-b sort plus bottom-up construction (default)\n");	\
    fprintf(stderr, "-n naive method\n");				\
    fprintf(stderr, "-s sort plus naive method\n");			\
  } while(0)

#define B_METHOD 'b'
#define N_METHOD 'n'
#define S_METHOD 's'

int main(int argc, char *argv[])
{   
  int res = MY_SUCCESS;
  char method = B_METHOD;
  char *infile = NULL;
  char *outfile = NULL;

  /*** INPUT ARGUMENT ***/  
  for(int i = 1; i < argc; i++) {
    if(argv[i][0] == '-') {
      switch (argv[i][1]){
      case B_METHOD: case N_METHOD: case S_METHOD:
	method = argv[i][1];
	break; 
      case 'v':
	fprintf(stderr, "%s\tversion %s\n\n", PROGNAME, PROGVERS);
	return MY_SUCCESS;
      case '?': case 'h': default:
	PRINT_USAGE(argv[0]); return  MY_SUCCESS;
      }
    } else {
      if(infile == NULL) {infile = argv[i];}
      else if(outfile == NULL) {outfile = argv[i];}
      else {PRINT_USAGE(argv[0]); return  MY_SUCCESS;}
    }
  }
  if(infile == NULL) {PRINT_USAGE(argv[0]); return  MY_SUCCESS;}
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO
  printf("%s\t\tversion %s\n", PROGNAME, PROGVERS);
  printf("package\t\t%s\n", ZDD_PACKAGE);
  printf("input\t\t%s\n", infile);
  if(outfile != NULL) printf("output\t\t%s\n", outfile);
  if(method == B_METHOD) {
    printf("method\t\tsort+bottom-up\n");
  } else if(method == N_METHOD) {
    printf("method\t\tnaive\n");
  } else if(method == S_METHOD) {
    printf("method\t\tsort+naive\n");
  } else {
    exit(EXIT_FAILURE);
  }
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO*/

  my_initrand(); // initialize random number generator
  FILE *in = fopen(infile, "r");
  ENSURE_TRUE_MSG(in != NULL, "file open failed\n");

  /*** INITIALIZE ZDD MANAGER ***/
  size_t nrow, maxval, total_nent;
  res = getfileinfo(&nrow, &maxval, &total_nent, in);
  ENSURE_SUCCESS(res);
  // adjust CACHE_DEFAULT_SIZE if necessary.
  zdd_init(maxval, CACHE_DEFAULT_SIZE);

  /*** IMPORT ZDD ***/  
  my_zddp f;
  if(method == B_METHOD) {
    clock_t startingtime = my_clock();
    f = zimport_setfam(nrow, total_nent, in);
    ENSURE_TRUE_MSG(f != ZDD_NULL, "ZDD import failed\n");
    print_elapsed_time(startingtime, "IMPORT");
  } else if (method == N_METHOD) {
    clock_t startingtime = my_clock();
    f = naive_zimport_setfam(nrow, total_nent, in);
    ENSURE_TRUE_MSG(f != ZDD_NULL, "ZDD import failed\n");
    print_elapsed_time(startingtime, "IMPORT");    
  } else if (method == S_METHOD) {
    clock_t startingtime = my_clock();
    f = sort_naive_zimport_setfam(nrow, total_nent, in);
    ENSURE_TRUE_MSG(f != ZDD_NULL, "ZDD import failed\n");
    print_elapsed_time(startingtime, "IMPORT");    
  } else {
    fprintf(stderr, "error: invalid method\n");
    exit(EXIT_FAILURE);
  }
#if defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO
    printf("|zdd|\t\t%zu\n", zdd_size(f));
    printf("#sets\t\t%zu\n\n", zdd_card(f));
#endif /*defined(LOG_LEVEL) && LOG_LEVEL <= LOG_INFO*/


  /*** EXPORT ZDD ***/  
  if(outfile != NULL) {
    FILE *out = fopen(outfile, "w");
    clock_t startingtime = my_clock();
    zexport_kf(maxval, f, out);
    print_elapsed_time(startingtime, "EXPORT");
    fclose(out);
  }

  zdd_quit();
  my_freerand();
  fclose(in);

  return MY_SUCCESS;
}


